
# THIS FILE IS GENERATED FROM wofry SETUP.PY
short_version = '1.0.7'
version = '1.0.7'
full_version = '1.0.7.dev0+c41c0c0'
git_revision = 'c41c0c03c0456a8fbf229e9b9cb89036385580ff'
release = False

if not release:
    version = full_version
    short_version += ".dev"
